"""Command line interface for kg-chat."""

import logging
from pathlib import Path
from pprint import pprint
from typing import Union

import click

from kg_chat import __version__
from kg_chat.app import create_app
from kg_chat.implementations.duckdb_implementation import DuckDBImplementation
from kg_chat.implementations.neo4j_implementation import Neo4jImplementation
from kg_chat.main import KnowledgeGraphChat

__all__ = [
    "main",
]

logger = logging.getLogger(__name__)
database_options = click.option(
    "--database",
    "-d",
    type=click.Choice(["neo4j", "duckdb"], case_sensitive=False),
    help="Database to use.",
    default="duckdb",
)
data_dir_option = click.option(
    "--data-dir",
    type=click.Path(exists=True, file_okay=False, dir_okay=True),
    help="Directory containing the data.",
    required=True,
)


@click.group()
@click.option("-v", "--verbose", count=True)
@click.option("-q", "--quiet")
@click.version_option(__version__)
def main(verbose: int, quiet: bool):
    """
    CLI for kg-chat.

    :param verbose: Verbosity while running.
    :param quiet: Boolean to be quiet or verbose.
    """
    if verbose >= 2:
        logger.setLevel(level=logging.DEBUG)
    elif verbose == 1:
        logger.setLevel(level=logging.INFO)
    else:
        logger.setLevel(level=logging.WARNING)
    if quiet:
        logger.setLevel(level=logging.ERROR)


@main.command("import")
@database_options
@data_dir_option
def import_kg(database: str = "duckdb", data_dir: str = None):
    """Run the kg-chat's demo command."""
    if not data_dir:
        raise ValueError("Data directory is required. This typically contains the KGX tsv files.")
    if database == "neo4j":
        impl = Neo4jImplementation(data_dir=data_dir)
        impl.load_kg()
    elif database == "duckdb":
        impl = DuckDBImplementation(data_dir=data_dir)
        impl.load_kg()
    else:
        raise ValueError(f"Database {database} not supported.")


@main.command()
@data_dir_option
@database_options
def test_query(data_dir: Union[str, Path], database: str = "duckdb"):
    """Run the kg-chat's chat command."""
    if database == "neo4j":
        impl = Neo4jImplementation(data_dir=data_dir)
        query = "MATCH (n) RETURN n LIMIT 10"
        result = impl.execute_query(query)
        for record in result:
            print(record)
    elif database == "duckdb":
        impl = DuckDBImplementation(data_dir=data_dir)
        query = "SELECT * FROM nodes LIMIT 10"
        result = impl.execute_query(query)
        for record in result:
            print(record)
    else:
        raise ValueError(f"Database {database} not supported.")


@main.command()
@data_dir_option
@database_options
def show_schema(data_dir: Union[str, Path], database: str = "duckdb"):
    """Run the kg-chat's chat command."""
    if database == "neo4j":
        impl = Neo4jImplementation(data_dir=data_dir)
        impl.show_schema()
    elif database == "duckdb":
        impl = DuckDBImplementation(data_dir=data_dir)
        impl.show_schema()
    else:
        raise ValueError(f"Database {database} not supported.")


@main.command()
@database_options
@click.argument("query", type=str, required=True)
@data_dir_option
def qna(query: str, data_dir: Union[str, Path], database: str = "duckdb"):
    """Run the kg-chat's chat command."""
    if database == "neo4j":
        impl = Neo4jImplementation(data_dir=data_dir)
        response = impl.get_human_response(query, impl)
        pprint(response)
    elif database == "duckdb":
        impl = DuckDBImplementation(data_dir=data_dir)
        response = impl.get_human_response(query)
        pprint(response)
    else:
        raise ValueError(f"Database {database} not supported.")


@main.command("chat")
@data_dir_option
@database_options
def run_chat(data_dir: Union[str, Path], database: str = "duckdb"):
    """Run the kg-chat's chat command."""
    if database == "neo4j":
        impl = Neo4jImplementation(data_dir=data_dir)
        kgc = KnowledgeGraphChat(impl)
        kgc.chat()
    elif database == "duckdb":
        impl = DuckDBImplementation(data_dir=data_dir)
        kgc = KnowledgeGraphChat(impl)
        kgc.chat()
    else:
        raise ValueError(f"Database {database} not supported.")


@main.command("app")
@click.option("--debug", is_flag=True, help="Run the app in debug mode.")
@data_dir_option
@database_options
def run_app(
    data_dir: Union[str, Path],
    debug: bool = False,
    database: str = "duckdb",
):
    """Run the kg-chat's chat command."""
    if database == "neo4j":
        impl = Neo4jImplementation(data_dir=data_dir)
        kgc = KnowledgeGraphChat(impl)
    elif database == "duckdb":
        impl = DuckDBImplementation(data_dir=data_dir)
        kgc = KnowledgeGraphChat(impl)
    else:
        raise ValueError(f"Database {database} not supported.")

    app = create_app(kgc)
    # use_reloader=False to avoid running the app twice in debug mode
    app.run(debug=debug, use_reloader=False)


if __name__ == "__main__":
    main()



"""Dash based app for the KG Chatbot."""

import io
import json
import zipfile

import dash
from dash import dcc, html
from dash.dependencies import Input, Output, State

from kg_chat.main import KnowledgeGraphChat
from kg_chat.utils import extract_nodes_edges, visualize_kg

# Global variable to store the last structured result
LAST_STRUCTURED_RESULT = None


def create_app(kg_chatbot: KnowledgeGraphChat):
    """Create a Dash app for the KG Chatbot."""
    # Initialize the Dash app
    app = dash.Dash(__name__, title="KG Chatbot")

    # Define the layout of the app
    app.layout = html.Div(
        [
            dcc.Store(id="conversation-history"),
            # Main content area
            html.Div(
                id="main-content",
                children=[
                    html.H1("KG Chat", style={"textAlign": "center"}),  # Add title here
                    dcc.Loading(
                        id="loading",
                        type="default",
                        children=html.Iframe(
                            id="graph-visualization",
                        ),
                    ),
                ],
            ),
            # Chatbot interface at the bottom right corner
            html.Div(
                id="chat-interface",
                children=[
                    html.Div(id="chat-output", style={"overflowY": "auto", "maxHeight": "300px"}),
                    dcc.Input(id="user-input", type="text", placeholder="Enter your prompt...", n_submit=0),
                    html.Button("Submit", id="submit-button"),
                    html.Button("Reset", id="reset-button", style={"margin-left": "10px"}),
                    html.Button("Download", id="download-button", style={"margin-left": "10px"}),
                    dcc.Download(id="download-results"),
                ],
            ),
        ]
    )

    def update_output(submit_n_clicks, reset_n_clicks, n_submit, value, history, kg_chatbot):
        """Update the chat output and graph visualization based on user input."""
        ctx = dash.callback_context

        if not ctx.triggered:
            return "", "", history, ""

        button_id = ctx.triggered[0]["prop_id"].split(".")[0]

        if (button_id == "submit-button" and submit_n_clicks) or (button_id == "user-input" and n_submit):
            if history is None:
                history = []

            try:
                if "show me" in value.lower():
                    structured_response = kg_chatbot.get_structured_response(value)
                    history.append(f"User: {value}")
                    cleaned_result = structured_response.replace("```", "").replace("json\n", "").replace("\n", "")
                    structured_result = json.loads(cleaned_result)
                    global LAST_STRUCTURED_RESULT
                    LAST_STRUCTURED_RESULT = structured_result  # Store the structured result
                    nodes, edges = extract_nodes_edges(structured_result)
                    html_str = visualize_kg(nodes, edges, app=True)
                    if html_str:
                        return [html.P(line) for line in history], html_str, history, ""
                else:
                    human_response = kg_chatbot.get_human_response(value)
                    history.append(f"User: {value}")
                    history.append(f"Bot: {human_response}")

                return [html.P(line) for line in history], "", history, ""

            except json.JSONDecodeError:
                history.append(f"User: {value}")
                history.append("Bot: Sorry, I could not understand the response format.")
                return [html.P(line) for line in history], "", history, ""
            except Exception as e:
                history.append(f"User: {value}")
                history.append(f"Bot: An error occurred: {e}")
                return [html.P(line) for line in history], "", history, ""

        elif button_id == "reset-button" and reset_n_clicks:
            return "", "", [], ""

        return "", "", history, ""

    @app.callback(
        Output("download-results", "data"),
        [Input("download-button", "n_clicks")],
        [State("conversation-history", "data")],
        prevent_initial_call=True,
    )
    def download_results(n_clicks, history):
        if not history:
            return None

        # Extract nodes and edges from the last structured response in the history
        try:
            last_response = LAST_STRUCTURED_RESULT
            if not last_response:
                return None

            nodes, edges = extract_nodes_edges(last_response)
            # Convert nodes and edges to TSV strings
            nodes_tsv = "id\tcategory\tlabel\n" + "\n".join(
                [f"{node['id']}\t{node['category']}\t{node['label']}" for node in nodes]
            )
            edges_tsv = "subject\tpredicate\tobject\n" + "\n".join(
                [f"{edge['subject']['id']}\t{edge['predicate']}\t{edge['object']['id']}" for edge in edges]
            )

            # Create a BytesIO buffer to hold the zip file
            buffer = io.BytesIO()
            with zipfile.ZipFile(buffer, "w") as zf:
                # Write nodes.tsv
                zf.writestr("nodes.tsv", nodes_tsv.encode("utf-8"))

                # Write edges.tsv
                zf.writestr("edges.tsv", edges_tsv.encode("utf-8"))

            buffer.seek(0)
            return dcc.send_bytes(buffer.getvalue(), "kg-subset.zip")

        except Exception as e:
            print(f"Error generating download: {e}")
            return None

    # Wrapper function to pass the database instance to the callback
    @app.callback(
        [
            Output("chat-output", "children"),
            Output("graph-visualization", "srcDoc"),
            Output("conversation-history", "data"),
            Output("user-input", "value"),
        ],
        [Input("submit-button", "n_clicks"), Input("reset-button", "n_clicks"), Input("user-input", "n_submit")],
        [State("user-input", "value"), State("conversation-history", "data")],
    )
    def _wrapped_update_output(submit_n_clicks, reset_n_clicks, n_submit, value, history):
        return update_output(submit_n_clicks, reset_n_clicks, n_submit, value, history, kg_chatbot)

    return app



"""Implementation of the DatabaseInterface for Neo4j."""

import csv
import time
from pathlib import Path
from pprint import pprint
from typing import Union

from langchain_community.chains.graph_qa.cypher import GraphCypherQAChain
from langchain_community.graphs import Neo4jGraph
from langchain_openai import ChatOpenAI
from neo4j import GraphDatabase

from kg_chat.constants import (
    DATALOAD_BATCH_SIZE,
    NEO4J_PASSWORD,
    NEO4J_URI,
    NEO4J_USERNAME,
    OPEN_AI_MODEL,
    OPENAI_KEY,
)
from kg_chat.interface.database_interface import DatabaseInterface
from kg_chat.utils import structure_query


class Neo4jImplementation(DatabaseInterface):
    """Implementation of the DatabaseInterface for Neo4j."""

    def __init__(
        self,
        data_dir: Union[str, Path],
        uri: str = NEO4J_URI,
        username: str = NEO4J_USERNAME,
        password: str = NEO4J_PASSWORD,
    ):
        """Initialize the Neo4j database and the Langchain components."""
        if not data_dir:
            raise ValueError("Data directory is required. This typically contains the KGX tsv files.")
        self.driver = GraphDatabase.driver(uri, auth=(username, password))
        self.graph = Neo4jGraph(url=uri, username=username, password=password)
        self.llm = ChatOpenAI(model=OPEN_AI_MODEL, temperature=0, api_key=OPENAI_KEY)
        self.data_dir = Path(data_dir)

        self.chain = GraphCypherQAChain.from_llm(
            graph=self.graph,
            llm=self.llm,
            verbose=True,
            use_function_response=True,
            # function_response_system="Respond as a Data Scientist!",
            validate_cypher=True,
        )
        self.safe_mode = True

    def toggle_safe_mode(self, enabled: bool):
        """Toggle safe mode on or off."""
        self.safe_mode = enabled

    def is_safe_command(self, command: str) -> bool:
        """Check if a command is safe to execute."""
        unsafe_keywords = ["CREATE", "DELETE", "REMOVE", "SET", "MERGE", "DROP"]
        return not any(keyword in command.upper() for keyword in unsafe_keywords)

    def execute_unsafe_operation(self, operation: callable, *args, **kwargs):
        """Execute an unsafe operation."""
        original_safe_mode = self.safe_mode
        self.safe_mode = False
        try:
            result = operation(*args, **kwargs)
        finally:
            self.safe_mode = original_safe_mode
        return result

    def execute_query(self, query: str):
        """Execute a Cypher query against the Neo4j database."""
        if self.safe_mode and not self.is_safe_command(query):
            raise ValueError(f"Unsafe command detected: {query}")
        with self.driver.session() as session:
            result = session.read_transaction(lambda tx: list(tx.run(query)))
        return result

    def execute_query_using_langchain(self, query: str):
        """Execute a Cypher query against the Neo4j database using Langchain."""
        if self.safe_mode and not self.is_safe_command(query):
            raise ValueError(f"Unsafe command detected: {query}")
        result = self.graph.query(query)
        return result

    def clear_database(self):
        """Clear the Neo4j database."""

        def _clear_database():
            with self.driver.session() as session:
                session.write_transaction(self._clear_database_tx)

        return self.execute_unsafe_operation(_clear_database)

    @staticmethod
    def _clear_database_tx(tx):
        """Clear the Neo4j database using APOC."""
        tx.run(
            """
        CALL apoc.periodic.iterate(
            'MATCH (n) RETURN n',
            'DETACH DELETE n',
            {batchSize:1000, parallel:true}
        )
        """
        )

    def ensure_index(self):
        """Ensure that the index on :Node(id) exists."""

        def _ensure_index():
            with self.driver.session() as session:
                session.write_transaction(self._ensure_index_tx)

        return self.execute_unsafe_operation(_ensure_index)

    @staticmethod
    def _ensure_index_tx(tx):
        """Index existence check."""
        tx.run(
            """
            CALL apoc.schema.assert(
                {Node: [['id'], ['category'], ['label']]},
                {RELATIONSHIP: [['subject'], ['predicate'], ['object']]}
            )
            """
        )
        print("Indexes ensured using APOC.")

    def get_human_response(self, query: str):
        """Get a human response from the Neo4j database."""
        human_response = self.chain.invoke({"query": query})
        pprint(human_response["result"])
        return human_response["result"]

    def get_structured_response(self, query: str):
        """Get a structured response from the Neo4j database."""
        response = self.chain.invoke({"query": structure_query(query)})
        return response["result"]

    def create_edges(self, edges):
        """Create relationships between nodes."""

        def _create_edges():
            with self.driver.session() as session:
                session.write_transaction(self._create_edges_tx, edges)

        return self.execute_unsafe_operation(_create_edges)

    @staticmethod
    def _create_edges_tx(tx, edges):
        tx.run(
            """
            UNWIND $edges AS edge
            MATCH (a:Node {id: edge.subject})
            MATCH (b:Node {id: edge.object})
            CREATE (a)-[r:RELATIONSHIP {type: edge.predicate}]->(b)
        """,
            edges=edges,
        )

    def create_nodes(self, nodes):
        """Create nodes in the Neo4j database."""

        def _create_nodes():
            with self.driver.session() as session:
                session.write_transaction(self._create_nodes_tx, nodes)

        return self.execute_unsafe_operation(_create_nodes)

    @staticmethod
    def _create_nodes_tx(tx, nodes):
        tx.run(
            """
            UNWIND $nodes AS node
            CREATE (n:Node {id: node.id, category: node.category, label: node.label})
        """,
            nodes=nodes,
        )

    def show_schema(self):
        """Show the schema of the Neo4j database."""
        with self.driver.session() as session:
            result = session.read_transaction(lambda tx: list(tx.run("CALL db.schema.visualization()")))
            pprint(result)

    def load_kg(self, block_size: int = DATALOAD_BATCH_SIZE):
        """Load the Knowledge Graph into the Neo4j database."""
        nodes_filepath = self.data_dir / "nodes.tsv"
        edges_filepath = self.data_dir / "edges.tsv"

        def _load_kg():
            # Clear the existing database
            print("Clearing the existing database...")
            self.clear_database()
            print("Database cleared.")

            # Ensure indexes are created
            print("Ensuring indexes...")
            self.ensure_index()
            print("Indexes ensured.")

            # Import nodes in batches
            print("Starting to import nodes...")
            start_time = time.time()
            nodes_batch = []
            columns_of_interest = ["id", "category", "name"]

            with open(nodes_filepath, "r") as nodes_file:
                reader = csv.DictReader(nodes_file, delimiter="\t")
                node_batch_loaded = 0

                for row in reader:
                    node_id = row[columns_of_interest[0]]
                    node_category = row[columns_of_interest[1]]
                    node_label = row[columns_of_interest[2]]
                    nodes_batch.append({"id": node_id, "category": node_category, "label": node_label})
                    node_batch_loaded += 1

                    if len(nodes_batch) >= block_size:
                        self.create_nodes(nodes_batch)
                        nodes_batch = []

                if nodes_batch:
                    self.create_nodes(nodes_batch)

            elapsed_time_seconds = time.time() - start_time

            if elapsed_time_seconds >= 3600:
                elapsed_time_hours = elapsed_time_seconds / 3600
                print(f"Nodes import completed: {node_batch_loaded} nodes in {elapsed_time_hours:.2f} hours.")
            else:
                elapsed_time_minutes = elapsed_time_seconds / 60
                print(f"Nodes import completed: {node_batch_loaded} nodes in {elapsed_time_minutes:.2f} minutes.")

            # Import edges in batches
            print("Starting to import edges...")
            start_time = time.time()
            edges_batch = []
            edge_column_of_interest = ["subject", "predicate", "object"]

            with open(edges_filepath, "r") as edges_file:
                reader = csv.DictReader(edges_file, delimiter="\t")
                edge_batch_loaded = 0

                for row in reader:
                    subject = row[edge_column_of_interest[0]]
                    predicate = row[edge_column_of_interest[1]]
                    object = row[edge_column_of_interest[2]]
                    edges_batch.append({"subject": subject, "predicate": predicate, "object": object})
                    edge_batch_loaded += 1

                    if len(edges_batch) >= block_size / 2:
                        self.create_edges(edges_batch)
                        edges_batch = []

                if edges_batch:
                    self.create_edges(edges_batch)

            elapsed_time_seconds = time.time() - start_time
            if elapsed_time_seconds >= 3600:
                elapsed_time_hours = elapsed_time_seconds / 3600
                print(f"Edges import completed: {edge_batch_loaded} edges in {elapsed_time_hours:.2f} hours.")
            else:
                elapsed_time_minutes = elapsed_time_seconds / 60
                print(f"Edges import completed: {edge_batch_loaded} edges in {elapsed_time_minutes:.2f} minutes.")

            print("Import process finished.")

        return self.execute_unsafe_operation(_load_kg)

    def __del__(self):
        """Ensure the driver is closed when the object is destroyed."""
        if hasattr(self, "driver"):
            self.driver.close()



"""Implementations of the chatbot."""

from .duckdb_implementation import DuckDBImplementation
from .neo4j_implementation import Neo4jImplementation

__all__ = ["DuckDBImplementation", "Neo4jImplementation"]



"""Utility functions for the KG chatbot."""

import random
import webbrowser

from pyvis.network import Network

from kg_chat.constants import GRAPH_OUTPUT_DIR

PREFIX_COLOR_MAP = {}


def extract_nodes_edges(structured_result):
    """Extract nodes and edges from the structured result."""
    nodes = structured_result.get("nodes", [])
    edges = structured_result.get("edges", [])

    return nodes, edges


def structure_query(query: str):
    """Structure the query to request structured results."""
    if "show me" in query.lower():
        # Modify the query to request structured results
        structured_query = f"""
        {query}
        Please provide the result in JSON format with ALL nodes and ALL edges. Return JSON ONLY.
        IMPORTANT: I want nothing but pure JSON (verbose=False).
        Example: {{
            "nodes": [
                {{"label": "A", "id": "1", category: "biolink:Gene"}},
                {{"label": "B", "id": "2", category: "biolink:ChemicalSubstance"}},
                {{"label": "C", "id": "3", category: "biolink:Disease"}},
                ..and so on
            ],
            "edges": [
                {{"subject": {{"label": "A", "id": "1"}},
                "object": {{"label": "B", "id": "2"}},
                "predicate": "biolink:related_to"}},
                ..and so on
            ]
        }}
        """
    else:
        structured_query = query
    return structured_query


def generate_random_color():
    """Generate a random color."""
    while True:
        color = "#{:06x}".format(random.randint(0, 0xFFFFFF))
        if color not in PREFIX_COLOR_MAP.values():
            return color


def assign_color_to_prefix(curie):
    """Assign a color to a prefix."""
    prefix = curie.split(":")[0]
    if prefix not in PREFIX_COLOR_MAP:
        PREFIX_COLOR_MAP[prefix] = generate_random_color()
    # print(f"PREFIX_COLOR_MAP={PREFIX_COLOR_MAP}") # To debug
    return PREFIX_COLOR_MAP[prefix]


def visualize_kg(nodes, edges, app: bool = False):
    """Visualize the knowledge graph using pyvis."""
    # Create a PyVis network
    net = Network(
        notebook=False,
        cdn_resources="in_line",
        neighborhood_highlight=True,
        select_menu=True,
        filter_menu=True,
    )

    if not nodes:
        print("No nodes to display.")
        return

    # Track added nodes to avoid duplicates
    added_nodes = set()

    # Add nodes to the network
    for node in nodes:
        # based on prefix (node["id"].split(":")[0]) assign nodes to color

        net.add_node(
            node["id"],
            label=node["label"],
            category=node.get("category", ""),
            title=node.get("id", ""),
            color=assign_color_to_prefix(node["id"]),
        )
        added_nodes.add(node["id"])

    # Add edges to the network
    for edge in edges:
        subject = edge["subject"]["id"]
        object = edge["object"]["id"]

        # Ensure both subject and object nodes are added
        if subject not in added_nodes:
            net.add_node(
                subject,
                label=edge["subject"].get("label", ""),
                category=edge["subject"].get("category", ""),
                title=subject,
                color=assign_color_to_prefix(subject),
            )
            added_nodes.add(subject)
        if object not in added_nodes:
            net.add_node(
                object,
                label=edge["object"].get("label", ""),
                category=edge["object"].get("category", ""),
                title=object,
                color=assign_color_to_prefix(object),
            )
            added_nodes.add(object)

        # Add edge with title (predicate)
        net.add_edge(subject, object, title=edge.get("predicate"))

    # Generate and display the network
    html_file = str(GRAPH_OUTPUT_DIR / "knowledge_graph.html")
    net.show(html_file, notebook=False)

    # Open the generated HTML file in the default web browser
    webbrowser.open(html_file)
    if app:
        # Generate the HTML representation
        return net.generate_html()



"""kg-chat package."""

import importlib_metadata

try:
    __version__ = importlib_metadata.version(__name__)
except importlib_metadata.PackageNotFoundError:
    # package is not installed
    __version__ = "0.0.0"  # pragma: no cover
